#include "listfile.h"
#include <QtCore>
#include <QTextStream>

/*
 * This file handles the ShelXl lst file during and after the refinement
 *
  - Mark shifts in red if over a certain level
  - show Max. 2-theta
 *
 */

ListFileHandler::ListFileHandler(QString lstpath, QString lstname, QObject *parent) : QObject(parent)
{
  REL_RESTR_CARDS << "SAME"<< "SADI"<< "SIMU"<< "RIGU"<< "ISOR"<< "NCSY"<< "FLAT" << "DELU";
  QFileInfo fi;
  fi.setFile(lstpath, lstname);
  full_lstpath = fi.absoluteFilePath();
  my_lstname = &lstname;
  QFileSystemWatcher fw;
  fw.addPath(full_lstpath);
  success = false; // turns to true if refinement was a sucess.
  listfcontent = new QStringList;
  listfcontent->append(this->loadLst());
}


QStringList ListFileHandler::loadLst(){
  //! Reads the SHELXL lst file from disk
  //! and returns it as a QStringlist
  QFile file(full_lstpath);
  QStringList lstList;
  if (file.open(QFile::ReadOnly | QFile::Text)) {
    QTextStream textStream(&file);
    while (true) {
      QString line = textStream.readLine();
      if (line.isNull()) {
          emit lstFinished();
          break;
      }
      if (line.isEmpty()) {
        continue;
      } else {
        lstList.append(line);
      }
    }
  }
  file.close();
  return lstList;
}

QVector<QStringList> ListFileHandler::listfileparser() {
  /*!
  Gathers the residuals of the SHELXL lst file.
  It searches for the final cycle summary and then for " Disagreeable restraints".
  End is reached with ' Summary of restraints'.
  Ported from: https://github.com/dkratzert/DSR-db
  */
  if (listfcontent->isEmpty()) {
    QVector<QStringList> dummy;
    return dummy;
  }
  bool disag = false;
  bool final = false;
  QVector<QStringList> disargeelist;
  int num = 0;
  foreach(QString line, *listfcontent) {
    QStringList splitline;
    splitline = line.split(" ", QString::SkipEmptyParts);
    if (splitline.isEmpty()) {
      continue;
    }
    if (splitline[0].startsWith("Observed")) {
      continue;
    }
    if (line.startsWith(" Final Structure Factor")) {
      final = true;
    }
    if (final && line.startsWith(" Disagreeable restraints")) {
      disag = true;
      continue;
    }
    if (line.startsWith(" Summary of restraints")) {
      final = false;
      disag = false;
    }
    if (line.contains("RIGU")) {
      continue;  // leave out RIGU, it produces too many lines...
    };
    if (splitline.size() < 2) continue;
    if (disag) {
      // in this case, the desired line is found:
      disargeelist.append(lineformatter(splitline));
      num = num+1;
      if (num > 500) {
        // Cutting restraints list. Too many bad restraints...
        return disargeelist;
      }
    }
  }
  return disargeelist;
}


QStringList ListFileHandler::lineformatter(QStringList line) {
  /*!
  takes care of some extra things with different restraints. For example
  RIGU xy should be in one column.
  Ported from: https://github.com/dkratzert/DSR-db
  */
  int num = 0;
  int pos = 0;
  foreach(QString i, line) {
    i = i.replace('/', ' ');
    if (i[0].isLetter()){
      pos = num;
      break;
    }
    num += 1;
  }
  // remove the part symbol from e.g. F1_2a:
  /*
  num = 0;
  foreach (QString i, line) {
    line[num] = remove_partsymbol(i);
    num += 1;
  } */
  // joining columns without numbers:
  QStringList posline = line.mid(pos, -1);
  QString joinline = posline.join(" ");
  QStringList newline = line.mid(0, pos);
  line = newline;
  line.append(joinline);
  QString tline = line.join(" ");
  foreach (QString n, REL_RESTR_CARDS) {
    if (tline.contains(n, Qt::CaseInsensitive)) {
      // adding placeholders for empty fields:
      QStringList nline;
      nline.append("-");
      nline.append("-");
      nline.append(line);
      return nline;
    }
  }
  return line;
}


QString ListFileHandler::table_maker(QVector<QStringList> tabledata) {;
    /*!
    Builds a html table out of a datalist from the final
    cycle summary of a shelxl list file.
    */
    QStringList table;
    foreach(QStringList line, tabledata) {
      table.append(this->row(line));
    }
    QString header = QString(
        "<body style=\"font-family:SansSerif; font-size:12pt;\">"
        "<table>"
        "<tr>"
          "<td width=60% align=left>"
          "<b>List of most disagreeable restraints:</b> &nbsp;"
          "</td>"
          "<td width=40% align=right>"
          "</td>"
        "</tr>"
        "</table>");
    QString footer = "</body>";

    QString empty_data = QString(
    "<table border=0 cellpadding=0 cellspacing=6 width=100% >"
      "<tr>"
         "<td align='center'> Observed </td>"
         "<td align='center'> Target   </td>"
         "<td align='center'> Error    </td>"
         "<td align='center'> Sigma    </td>"
         "<td align='left'> Restraint  </td>"
      "</tr>"
        "<tr>"
         "<td align='center'> All fine... </td>"
         "<td align='center'> -- </td>"
         "<td align='center'> -- </td>"
         "<td align='center'> -- </td>"
         "<td align='center'> -- </td>"
      "</tr>"
    "</table>");

    QString html = QString(
      "%1"
      "<table border=0 cellpadding=0 cellspacing=6 width=100% >"
        "<tr>"
           "<td align='center'> Observed </td>"
           "<td align='center'> Target   </td>"
           "<td align='center'> Error    </td>"
           "<td align='center'> Sigma    </td>"
           "<td align='left'> Restraint  </td>"
        "</tr>"
        "%2"
      "</table>"
        "%3").arg(header, table.join("\n"), footer);

    if(table.isEmpty()) {
      return header+empty_data;
    }
    return html;
}

QString ListFileHandler::row(QStringList rowdata) {
    /*!
    creates a table row for the restraints list.
    :type rowdata: list
    */
    QStringList td;
    td.clear();
    QString row = "";
    QString bgcolor = "";
    if (rowdata.size() >= 5) {
      bool ok1 = false;
      bool ok2 = false;
      rowdata[2].toFloat(&ok1);
      rowdata[3].toFloat(&ok2);
      if (ok1 && ok2) {
        if (qAbs(rowdata.at(2).toFloat()) > 2.5*rowdata[3].toFloat()) {
          bgcolor = QString("bgcolor='#ffec99'"); // bgcolor='#FFD100'  // yellow
        }
        if (qAbs(rowdata.at(2).toFloat()) > 3.5*rowdata[3].toFloat()) {
          bgcolor = QString("bgcolor='#ff9fac'"); // bgcolor='#FF1030'  // red
        }
      }
    }
    int num = 0;
    foreach (QString item, rowdata) {
      bool ok = false;
      item.toDouble(&ok);
      if (ok) {
        // align right for numbers:
        if (num < 2) {
          // do not colorize the first two columns:
          td.append(QString("<td align='right'> %1 </td>").arg(item));
        } else {
          td.append(QString("<td align='right' %1> %2 </td>").arg(bgcolor, item));
        }
      } else {
        if (item.startsWith('-')) {
          // only a minus sign
          td.append(QString("<td align='center'> %1 </td>").arg(item));
        } else {
          if (num < 4) {
            td.append(QString("<td align='right'> %1 </td>").arg(item));
            continue;
          }
          // align left for words:
          td.append(QString("<td align='left'> %1 </td>").arg(item));
        }
      }
      num += 1;
    }
    if (td.isEmpty()) {
      row = "<tr> No (disagreeable) restraints found in .lst file. </tr>";
    } else {
      row = QString("<tr> %1 </tr>").arg(td.join(""));
    }
    td.clear();
    return row;
}

QString ListFileHandler::restraintsTable() {
  return table_maker(listfileparser());
}

bool ListFileHandler::getSuccess() const {
    return success;
  }

void ListFileHandler::setSuccess(bool value) {
    success = value;
  }

QString ListFileHandler::thetaFromLst() {
  //! Max. 2-theta =   55.30
  //!   -8 =< h =<  8,    -21 =< k =< 21,    -21 =< l =< 21,   Max. 2-theta =   55.30
  //! I might need this later
  QString theta = "";
  foreach(QString line, *listfcontent) {
    if (line.contains("Max. 2-theta")) {
      theta = line.section("theta", 1, 1).split(" ", QString::SkipEmptyParts).at(1);
    }
  }
  return theta;
}


// ##########################################################################


XLOutputParser::XLOutputParser(QObject *parent) : QObject(parent)
{
  //! Handles the content of the SHELXL list file.
  //! Methods with "current" in the name return values of the current state
  //! in the continuously growing SHELXL list file.
  //! Methods with "final" return the values after successful refinement.
  //!
  // The constructor:
  listf_data.cmdline = "";
  listf_data.xlversion = "";
  listf_data.threads = "";
  listf_data.procs = "";
  listf_data.rint = "";
  listf_data.rsigma = "";
  listf_data.wr2 = "";
  listf_data.cycle = "";
  listf_data.data = "";
  listf_data.current_parameters = "";
  listf_data.all_parameters = "";
  listf_data.curGoof = " ";
  listf_data.curRGoof = " ";
  listf_data.finalR1 = "";
  listf_data.finalwR2 = "";
  listf_data.maxshift_esd = "";
  listf_data.maxshiftfor = "";
  listf_data.meanshift = "";
  listf_data.maxshift = "";
  listf_data.dU = "";
  listf_data.dUfor = "";
  listf_data.peak = "";
  listf_data.hole = "";
  listf_data.peakpos = "";
  listf_data.holepos = "";
  listf_data.peakdist = "";
  listf_data.holedist = "";
  listf_data.warning = "";
  listf_data.warnings.clear();

  rgr.Goof = 0.0;
  rgr.cycle = 0.0;
  rgr.dU = 0.0;
  rgr.dUfor = "";
  rgr.wR2 = 0.0;
  rgr.final_wR2 = 0.0;
  rgr.max_shift_esd_for = "";
  rgr.maxshift = 0.0;
  rgr.maxshift_esd = 0.0;
  rgr.maxshiftfor = "";
  rgr.meanshift = 0.0;
  rgr.warnings.clear();
  warnings.clear();
}


RGraph XLOutputParser::residualsGraphDisplay() {
  //! Returns a data structure that collects the values of
  //! wR2, R1, Goof, ... during refinement to display them as
  //! a graph.
  // collect data from each line and append it to RGraph
  rgr.Goof = listf_data.curGoof.toDouble();
  rgr.wR2 = listf_data.wr2.toDouble();
  rgr.maxshift = listf_data.maxshift.toDouble();
  rgr.meanshift = listf_data.meanshift.toDouble();
  rgr.cycle = listf_data.cycle.toInt();
  rgr.final_wR2 = listf_data.finalwR2.toDouble();
  rgr.warnings = listf_data.warnings;
  return rgr;
}


void XLOutputParser::fillListFileData(QString line) {
  //! Fills listf_data with data but it is only filled
  //! if the respective strings contain data
  QStringList tproc = xl_threads(line);
  QString cmdline = xl_cmdline(line);
  QString version = xl_version(line);
  QStringList rintsig = xl_rint_rsigma(line);
  QStringList wr2data = xl_current_wr2_data_param(line);
  QString finalR1 = xl_final_R1(line);
  QString finalwR2 = xl_final_wR2(line);
  QStringList goofs = xl_current_goof(line);
  QStringList peakvalues = xl_peak(line);
  QStringList holevalues = xl_hole(line);
  //QStringList fgoofs = xl_final_goofs(line);
  QStringList maxshifts = xl_current_max_shift(line);
  QStringList meanshifts = xl_current_meanshift_esd(line);
  xl_collect_warnings(line);
  QStringList flackvals = xl_flacksvals(line);

  // Only fill listf_data with values if there are any:
  if (!cmdline.isEmpty()) {
    listf_data.cmdline = cmdline;
  }
  if (!version.isEmpty()) {
    listf_data.xlversion = version;
  }
  if ((!tproc[0].isEmpty()) && (!tproc[1].isEmpty())) {
    listf_data.threads = tproc[0];
    listf_data.procs = tproc[1];
  }
  if ((!rintsig[0].isEmpty()) && (!rintsig[1].isEmpty())) {
    listf_data.rint = rintsig[0];
    listf_data.rsigma = rintsig[1];
  }
  if (!wr2data.at(0).isEmpty()) {
    listf_data.wr2 = wr2data.at(0);
  }
  if (!wr2data.at(1).isEmpty()) {
    emit newLScycle();
    listf_data.cycle = wr2data.at(1);
  }
  if (!wr2data.at(2).isEmpty()) {
    listf_data.data = wr2data.at(2);
  }
  if (!wr2data.at(3).isEmpty()) {
    listf_data.current_parameters = wr2data.at(3);
  }
  if (!wr2data.at(4).isEmpty()) {
    listf_data.all_parameters = wr2data.at(4);
  }
  if (!goofs[0].isEmpty() && !goofs[1].isEmpty()) {
    listf_data.curGoof = goofs[0];
    listf_data.curRGoof = goofs[1];
  }
  if (!finalR1.isEmpty()) {
    listf_data.finalR1 = finalR1;
  }
  if (!finalwR2.isEmpty()) {
    listf_data.finalwR2 = finalwR2;
  }
  if (!maxshifts[0].isEmpty() && !maxshifts[1].isEmpty()) {
    //shifts.append(maxshift);
    //shifts.append(maxshiftfor);
    //shifts.append(dU);
    //shifts.append(dUfor);
    listf_data.maxshift = maxshifts[0];
    listf_data.maxshiftfor = maxshifts[1];
    listf_data.dU = maxshifts[2];
    listf_data.dUfor = maxshifts[3];
  }
  if (!meanshifts[0].isEmpty() && !meanshifts[1].isEmpty()) {
    //shifts.append(meanshift);
    //shifts.append(maxshift_esd);
    //shifts.append(max_esd_shiftfor);
    listf_data.meanshift = meanshifts[0];
    listf_data.maxshift_esd = meanshifts[1];
    listf_data.max_shift_esd_for = meanshifts[1];
  }
  if (!holevalues[0].isEmpty() && !holevalues[1].isEmpty() && !holevalues[2].isEmpty()) {
    listf_data.hole = holevalues[0];
    listf_data.holepos = holevalues[1];
    listf_data.holedist = holevalues[2];
  }
  if (!peakvalues[0].isEmpty() && !peakvalues[1].isEmpty() && !peakvalues[2].isEmpty()) {
    listf_data.peak = peakvalues[0];
    listf_data.peakpos = peakvalues[1];
    listf_data.peakdist = peakvalues[2];
  }
  if (!flackvals[0].isEmpty()) {
    listf_data.flack = flackvals.at(0);
  }
  if (!flackvals[1].isEmpty()) {
    listf_data.parson = flackvals.at(1);
  }
}

QString XLOutputParser::residualsTextDisplay(QString line){
  //! returns a html string to display the refinement results in a QLabel
  fillListFileData(line);
  QString html_final;
  QString wr2line;
  QString shiftyellow = "";
  QString dUyellow = "";
  if (!listf_data.finalR1.isEmpty() && !listf_data.finalwR2.isEmpty()) { // final values:
    html_final = QString("<tr><td align=right><b>Final <b><i>R</i>1</b> =    </td> <td> %1%     </td>"
                         "    <td align=right><b><i>wR</i><sub>2</sub></b> = </td> <td> %2%     </td></tr>"
                         "<tr><td align=right><b>peak</b> =          </td> <td> %3 e&#8491;<sup>-3</sup> at %4 </td>"
                         "    <td align=right><b>hole</b> =          </td> <td> %5 e&#8491;<sup>-3</sup> at %6 </td></tr>"
                         )
          .arg(listf_data.finalR1.toFloat()*100, 10, 'f', 2)      // 1
          .arg(listf_data.finalwR2.toFloat()*100, 10, 'f', 2)     // 2
          .arg(listf_data.peak)           // 3
          .arg(listf_data.peakdist)       // 4
          .arg(listf_data.hole)           // 5
          .arg(listf_data.holedist)       // 6
        ;
    if (!listf_data.flack.isEmpty() || !listf_data.parson.isEmpty()) {
      wr2line = QString("<tr><td align=right> Flack X  = </td> <td> %1 </td> "
                           "<td align=right> Parson Z = </td> <td> %2 </td></tr>").arg(listf_data.flack, listf_data.parson);
    } else { // no flack parameter:
      wr2line = "<tr><td> </td></tr>";
    }
  } else {  // shelxl is currently runnig:
    html_final = "";
    wr2line = QString("<tr><td align=right> <i>wR</i><sub>2</sub> =   </td> <td> %7 % </td></tr>")
                      .arg(listf_data.wr2.toFloat()*100, 10, 'f', 2);
  }
  if (qAbs(listf_data.maxshift.toFloat()) > 0.005) {
    shiftyellow = QString("bgcolor='#ffec99'");
  }
  if (qAbs(listf_data.dU.toFloat()) > 0.005) {
    dUyellow = QString("bgcolor='#ffec99'");
  }
  QString html_output = QString(
          "<H3>Running SHELXL %1</H3> "
          "<body style=\"font-family:SansSerif; font-size:12pt;\">"
          ""
          "<table cellpadding=2%>"
          "<tr><td>%2 threads on %3 processors </td></tr>"
          "<tr><td>Using %9 data and %12 of %10 parameters</td> <td> (ratio %18) &nbsp; </td>"
          "<td><i>R</i><sub>int</sub> = %5%, <i>R</i><sub>&sigma;</sub> = %6% </td></tr>"
          "</table>"
          "<table cellpadding=2% width=100% align=left>"
          "<tr><td width=14%> <b>Cycle %8 </b>            </td></tr>"
          "%7"
          "<tr><td align=right> <i>Goof</i> =             </td> <td> %11                </td>"
          "    <td align=right  width=18%> Restrained <i>Goof</i> =  </td> <td> %19                </td></tr>"
          "<tr><td align=right> Max. shift =              </td> <td %4> %14 &#8491; - %15  </td>"
          "    <td align=right> Max. &Delta;U =           </td> <td %21> %16 - %17          </td></tr>"
          "%13"
          "</table>"
          "<table><tr><td bgcolor='#ffec99'> %20 </td></tr></table>"
          "</body>")
          .arg(listf_data.xlversion)       // 1
          .arg(listf_data.threads)         // 2
          .arg(listf_data.procs)           // 3
          .arg(shiftyellow)                // 4
          .arg(listf_data.rint.toFloat()*100, 10, 'f', 2)         // 5
          .arg(listf_data.rsigma.toFloat()*100, 10, 'f', 2)       // 6
          .arg(wr2line)                         // 7 //wr2
          .arg(listf_data.cycle)                // 8
          .arg(listf_data.data)                 // 9
          .arg(listf_data.all_parameters)       // 10
          .arg(listf_data.curGoof)              // 11
          .arg(listf_data.current_parameters)   // 12
          .arg(html_final)                // 13
          .arg(listf_data.maxshift)       // 14
          .arg(listf_data.maxshiftfor)    // 15
          .arg(listf_data.dU)             // 16
          .arg(listf_data.dUfor)          // 17
          .arg(listf_data.data.toFloat()/listf_data.all_parameters.toFloat(), 10, 'f', 1) // 18
          .arg(listf_data.curRGoof)       // 19
          .arg(listf_data.warnings.join("<br>")) // 20
          .arg(dUyellow)                         // 21

      ;
  return html_output;
}


QString XLOutputParser::xl_version(QString line) {
  //! Returns the running shelxl version
  //! " +  Copyright(C) George M. Sheldrick 1993-2014     Version 2014/7  +"
  QRegExp versionregex = QRegExp("Version\\s+\\d+\\/\\d+");
  QString version;
  QStringList linesplit;
  if (line.contains(versionregex)) {
    linesplit = line.split(" ", QString::SkipEmptyParts);
    int index = linesplit.indexOf("Version");
    version = linesplit.at(index+1);
  }
  return version;
}

QStringList XLOutputParser::xl_threads(QString line) {
  //! Returns on how many threads and processors shelxl is running
  //! "Running  6 threads on  8 processors"
  QStringList threadproc;
  QString threads = "";
  QString processors = "";
  QStringList linesplit;
  QRegExp tprocregex = QRegExp("Running\\s+\\d\\s+threads\\s+on\\s+\\d\\s+processors");
  if (line.contains(tprocregex)) {
    linesplit = line.split(" ", QString::SkipEmptyParts);
    threads = linesplit.at(1);
    processors = linesplit.at(4);
  }
  threadproc.append(threads);
  threadproc.append(processors);
  return threadproc;
}

QString XLOutputParser::xl_cmdline(QString line) {
  //! Returns the command line options shelxl was started with
  //! " Command line parameters: p21c -a50000 -b3000 -c624 -t6"
  QString cmdline;
  if (line.contains("Command line parameters:")) {
    cmdline = line;
  }
  return cmdline;
}

QStringList XLOutputParser::xl_current_wr2_data_param(QString line) {
  //! Returns the number of data (reflections)
  //! shelxl is refinimg against. This is without sigma cutoff.
  //! " wR2 = 0.4016 before cycle   1 for   10786 data and    737 /    737 parameters"
  QStringList linesplit;
  QString wr2 = "";
  QString cycle = "";
  QString data = "";
  QString current_parameters = "";
  QString all_parameters = "";
  QStringList alldata;
  if (line.contains("wR2") && line.contains("cycle") && line.contains("data")
      && (line.split(" ", QString::SkipEmptyParts).size() >= 12) ) {
    linesplit = line.split(" ", QString::SkipEmptyParts);
    wr2 = linesplit.at(2);
    cycle = linesplit.at(5);
    data = linesplit.at(7);
    current_parameters = linesplit.at(10);
    all_parameters = linesplit.at(12);
  }
  alldata.append(wr2);
  alldata.append(cycle);
  alldata.append(data);
  alldata.append(current_parameters);
  alldata.append(all_parameters);
  return alldata;
}

QStringList XLOutputParser::xl_rint_rsigma(QString line) {
  //! Returns Rint and Rsigma of the dataset.
  //! " R(int) = 0.0504     R(sigma) = 0.0585      Friedel opposites merged"
  QString rint = "";
  QString rsigma = "";
  QStringList linesplit, rvals;
  if (line.contains("R(int)") && (line.split(" ", QString::SkipEmptyParts).size() >= 9) ) {
    linesplit = line.split(" ", QString::SkipEmptyParts);
    rint = linesplit.at(7);
    rsigma = linesplit.at(10);
  }
  rvals.append(rint);
  rvals.append(rsigma);
  return rvals;
}

int XLOutputParser::xl_syst_abs_bad_equiv(QString line) {
  //! Returns the number of bad equivalents reported by shelxl.
  int badequiv = 3;
  qDebug() << line;
  return badequiv;
}

QStringList XLOutputParser::xl_current_goof(QString line) {
  //! Returns the current Goof of the refinement. This value is
  //! the last value in the continuesly growing lst file.
  //! "GooF = S =     4.130;     Restrained GooF =      5.432 for    1662 restraints"
  QString goof;
  QString rgoof;
  QStringList linesplit;
  QStringList goofs;
  if (line.contains("Restrained GooF") && line.contains("restraints")
      && (line.split(" ", QString::SkipEmptyParts).size() >= 8)) {
    linesplit = line.split(" ", QString::SkipEmptyParts);
    goof = linesplit.at(4);
    goof.remove(QChar(';'));
    rgoof = linesplit.at(8);
  }
  goofs.append(goof);
  goofs.append(rgoof);
  return goofs;
}

QStringList XLOutputParser::xl_final_goofs(QString line) {
  //! Returns the final Goof after refinement finished.
  //! "wR2 =  0.2277,  GooF = S =   2.275,  Restrained GooF =    2.122  for all data"
  QString goof;
  QString rgoof;
  QStringList linesplit;
  QStringList goofs;
  if (line.contains("for all data") && (line.split(" ", QString::SkipEmptyParts).size() >= 11) ) {
    linesplit = line.split(" ", QString::SkipEmptyParts);
    goof = linesplit.at(7);
    goof.remove(QChar(','));
    rgoof = linesplit.at(11);
  }
  goofs.append(goof);
  goofs.append(rgoof);
  return goofs;
}


QStringList XLOutputParser::xl_current_meanshift_esd(QString line) {
  //! Returns the current shift per esd.
  //! "Mean shift/esd =   0.016  Maximum =    -8.976 for  U11 C6         at 09:36:30
  QStringList shifts;
  QString meanshift;
  QString maxshift_esd;
  QString max_esd_shiftfor;
  QStringList linesplit;
  if (line.contains("shift/esd") && line.contains("Maximum")
      && (line.split(" ", QString::SkipEmptyParts).size() >= 6) ) {
    linesplit = line.split(" ", QString::SkipEmptyParts);
    meanshift = linesplit.at(3);
    maxshift_esd = linesplit.at(6);
    max_esd_shiftfor = line.split("for", QString::SkipEmptyParts).at(1).trimmed()
        .split("at", QString::SkipEmptyParts).at(0).trimmed();
  }
  shifts.append(meanshift);
  shifts.append(maxshift_esd);
  shifts.append(max_esd_shiftfor);
  return shifts;
}


QStringList XLOutputParser::xl_current_max_shift(QString line) {
  //! Returns the maximum shift during the refinement.
  //! "Max. shift = 0.000 A for H26C      Max. dU =-0.007 for C6"
  //! "=-0.007 for C6"
  QString maxshift;
  QString maxshiftfor;
  QString dU;
  QString dUfor;
  QStringList linesplit;
  QStringList shifts;
  if (line.contains("Max. shift") && line.contains("dU") && line.contains("for")
      && (line.split(" ", QString::SkipEmptyParts).size() >= 6) ) {
    linesplit = line.split(" ", QString::SkipEmptyParts);
    maxshift = linesplit.at(3);
    maxshiftfor = linesplit.at(6);
    dU = line.split("dU").at(1).trimmed().remove("=").split(" ", QString::SkipEmptyParts).at(0);
    dUfor = line.split("for").at(2).trimmed();
  }
  shifts.append(maxshift);
  shifts.append(maxshiftfor);
  shifts.append(dU);
  shifts.append(dUfor);
  return shifts;
}


QString XLOutputParser::xl_current_warning(QString line) {
  //! Returns a Warning if there is any in the current line.
  QRegExp warnregexp = QRegExp("\\*\\*.*\\*\\*");
  QString warn = "";
  if ( line.contains(warnregexp) ) {
    warn = line.simplified();
  }
  return warn;
}

void XLOutputParser::xl_collect_warnings(QString line) {
  //! Returns all warning that occoured during refinement.
  QRegExp warnregexp = QRegExp("\\*\\*.*\\*\\*");
  if ( line.contains(warnregexp) ) {
    warnings.append(line.simplified());
  }
  if (line.contains("CANNOT OPEN FILE")) {
    warnings.append(line.simplified());
  }
  listf_data.warnings = warnings;
  rgr.warnings = warnings;
}

QString XLOutputParser::xl_final_R1(QString line) {
  QString finalr1 = "";
  QStringList linesplit;
  if (line.contains("R1") && line.contains("4sig") && line.contains("for all")) {
    linesplit = line.split(" ", QString::SkipEmptyParts);
    finalr1 = linesplit.at(2);
    finalr1.remove(QChar(','));
  }
  return finalr1;
}

QString XLOutputParser::xl_final_wR2(QString line) {
  QString finalwr2;
  QStringList linesplit;
  if (line.contains("wR2") && line.contains("for all data")) {
    linesplit = line.split(" ", QString::SkipEmptyParts);
    finalwr2 = linesplit.at(2);
    finalwr2.remove(QChar(','));
  }
  return finalwr2;
}

QStringList XLOutputParser::xl_peak(QString line) {
  //! Returns a list of the highest peak and the position
  // Highest peak    8.15  at  0.3160  0.7496  0.4960  [  1.30 A from F12 ]
  QStringList values;
  QString peak = "";
  QString coord = "";
  QString dist = "";
  QStringList linesplit;
  if (line.contains("Highest peak") && line.contains("from")
      && (line.split(" ", QString::SkipEmptyParts).size() >= 6) ) {
    linesplit = line.split(" ", QString::SkipEmptyParts);
    peak = linesplit.at(2);
    coord = linesplit.at(4)+" "+linesplit.at(5)+" "+linesplit.at(6);
    dist = line.split("[").at(1).split("]").at(0).trimmed();
  }
  values.append(peak);
  values.append(coord);
  values.append(dist);
  return values;
}

QStringList XLOutputParser::xl_hole(QString line) {
  //! Returns a list of the deepest hole and the position
  // Deepest hole   -0.74  at  0.4337  0.4402  0.7821  [  0.84 A from GA1 ]
  QStringList values;
  QString hole = "";
  QString coord = "";
  QString dist = "";
  QStringList linesplit;
  if (line.contains("Deepest hole") && line.contains("from")
      && (line.split(" ", QString::SkipEmptyParts).size() >= 12) ) {
    linesplit = line.split(" ", QString::SkipEmptyParts);
    hole = linesplit.at(2);
    coord = linesplit.at(4)+" "+linesplit.at(5)+" "+linesplit.at(6);
    dist = line.split("[").at(1).split("]").at(0).trimmed();
  }
  values.append(hole);
  values.append(coord);
  values.append(dist);
  return values;
}

QStringList XLOutputParser::xl_flacksvals(QString line) {
  //! Flack x =   -0.449(792) by classical fit to all intensities
  //! Flack x =   -0.315(296) from 3107 selected quotients (Parsons' method)
  QStringList values;
  QString flack = "";
  QString parson = "";
  QStringList linesplit;
  if (line.contains("Flack x") && line.contains("classical fit")
      && (line.split(" ", QString::SkipEmptyParts).size() >= 3) ) {
    linesplit = line.split(" ", QString::SkipEmptyParts);
    flack = linesplit.at(3);
  }
  if (line.contains("Flack x") && line.contains("Parsons' method")
      && (line.split(" ", QString::SkipEmptyParts).size() >= 3) ) {
    linesplit = line.split(" ", QString::SkipEmptyParts);
    parson = linesplit.at(3);

  }
  values.append(flack);
  values.append(parson);
  return values;
}


